# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""The module for classes
"""

import math
import torch

class ResumableSequentialSampler:
    """ The resumble sequential sampler.
    """
    def __init__(self,
                 samples_per_epoch: int,
                 batch_size: int,
                 drop_last: bool = False,
                 data_parallel_rank: int = 0,
                 data_parallel_size: int = 1):
        self._samples_per_epoch = samples_per_epoch
        self._batch_size = batch_size
        self._drop_last = drop_last
        self._consumed_samples = 0
        self._cur_index = 0

        assert batch_size % data_parallel_size == 0, "batch_size should be a multiple of data_parallel_size"
        assert samples_per_epoch % data_parallel_size == 0 or drop_last is True, \
            "ResumableSampler does not support situation where " + \
            "samples_per_epoch is not divisible by data_parallel_size and drop_last is not set " + \
            "since it may cause undesirable sample duplicates. "

        self._data_parallel_rank = data_parallel_rank
        self._data_parallel_size = data_parallel_size

    def set_processed_steps(self, processed_steps: int):
        steps_per_epoch = math.ceil(self._samples_per_epoch / self._batch_size)
        self._consumed_samples = (processed_steps // steps_per_epoch) * self._samples_per_epoch + \
            (processed_steps % steps_per_epoch) * self._batch_size
        self._cur_index = self._consumed_samples % self._samples_per_epoch

    def __len__(self):
        return self._samples_per_epoch

    def __iter__(self):
        batch = []
        for index in range(self._cur_index, self._samples_per_epoch):
            batch.append(index)
            if len(batch) == self._batch_size:
                yield batch[self._data_parallel_rank * (self._batch_size // self._data_parallel_size) \
                    :(self._data_parallel_rank + 1) * (self._batch_size // self._data_parallel_size)]
                batch = []
        if len(batch) > 0 and not self._drop_last:
            yield batch[self._data_parallel_rank * (len(batch) // self._data_parallel_size) \
                    :(self._data_parallel_rank + 1) * (len(batch) // self._data_parallel_size)]
            batch = []
        # Initialize the cur index for the future epochs
        self._cur_index = 0


class ResumableRandomSampler:
    """ The random sampler that supports deterministic & resumable random sampling.
    """
    def __init__(self,
                 samples_per_epoch: int,
                 batch_size: int,
                 drop_last: bool = False,
                 seed: int = 0,
                 data_parallel_rank: int = 0,
                 data_parallel_size: int = 1):
        self._samples_per_epoch = samples_per_epoch
        self._batch_size = batch_size
        self._cur_epoch = 0
        self._consumed_samples_cur_epoch = 0
        self._drop_last = drop_last
        self._seed = seed

        assert batch_size % data_parallel_size == 0, "batch_size should be a multiple of data_parallel_size"
        assert samples_per_epoch % data_parallel_size == 0 or drop_last is True, \
            "ResumableRandomSampler does not support situation where" + \
            "samples_per_epoch is not divisible by data_parallel_size and drop_last is not set" + \
            "since it may cause undesirable sample duplicates."

        self._data_parallel_rank = data_parallel_rank
        self._data_parallel_size = data_parallel_size

    def set_processed_steps(self, processed_steps: int):
        steps_per_epoch = math.ceil(self._samples_per_epoch / self._batch_size)
        self._cur_epoch = processed_steps // steps_per_epoch
        self._consumed_samples_cur_epoch = (processed_steps % steps_per_epoch) * self._batch_size

    def __len__(self):
        return self._samples_per_epoch

    def __iter__(self):
        g = torch.Generator()
        # We avoid seed 0 for good random generation.
        g.manual_seed(self._cur_epoch + self._seed + 1)
        random_indices = torch.randperm(self._samples_per_epoch // self._data_parallel_size, generator=g).tolist()
        start_idx = self._consumed_samples_cur_epoch // self._data_parallel_size
        offset = (self._samples_per_epoch // self._data_parallel_size) * self._data_parallel_rank
        batch = []
        for index in range(start_idx, self._samples_per_epoch // self._data_parallel_size):
            batch.append(offset + random_indices[index])
            if len(batch) == self._batch_size // self._data_parallel_size:
                yield batch
                batch = []
        if len(batch) > 0 and not self._drop_last:
            yield batch
        # Advance one epoch for future reuses
        self._cur_epoch += 1
        self._consumed_samples_cur_epoch = 0



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow training manager module.
"""

import atexit
import json
import logging
import os
import time
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from threading import Thread
from typing import Dict, Optional

import torch

from periflow_sdk.comm.ipc import (
    IpcCommPurpose,
    get_default_ipc_channel,
    IpcChannel,
    ipc_read,
    ipc_status_read,
    ipc_write
)
from periflow_sdk.errors import PeriFlowError, PeriFlowInternalError
from periflow_sdk.utils import (
    check_initialized,
    ensure_divisibility,
    SaveType
)


periflow_logger = logging.getLogger("PERIFLOW_SDK")


class TrainingManager:

    """ The training wrapper class for general PyTorch training code.
    """
    def __init__(self, teardown_at_exit: bool = True):
        self.has_initialized: bool = False

        self._is_local: bool = os.environ.get("PERIFLOW_ENABLED", "0") != "1"

        self._cur_step: int = 0
        self._step_start_time: Optional[float] = None

        self._local_rank: Optional[int] = None
        self._rank: Optional[int] = None

        # IPC Channels
        self._ipc_channels: Dict[IpcCommPurpose, IpcChannel] = {}

        # Emergency save
        self._wait_emergency_save_thread: Optional[Thread] = None
        self._emergency_save_step: Optional[int] = None

        # Used only for local mode
        self._log_path: Optional[Path] = None

        # Checkpoint tracking
        self._checkpoint_saved_cur_step: bool = False

        if not self._is_local and teardown_at_exit:
            atexit.register(self._teardown)

    def _cloud_init(self):
        """post-init for cloud mode
        """
        # Environment variable check.
        required_env_vars = ["WORLD_SIZE",
                             "NODE_RANK",
                             "NUM_NODES",
                             "PROCESSED_ITERS"]

        for env_var in required_env_vars:
            if env_var not in os.environ:
                raise PeriFlowInternalError(
                    f"Environment variable '{env_var}' should be set in cloud mode!"
                )

        # Configure dist info
        world_size = int(os.environ["WORLD_SIZE"])
        num_nodes = int(os.environ["NUM_NODES"])
        ensure_divisibility(world_size, num_nodes)
        devices_per_node = world_size // num_nodes

        if torch.distributed.is_initialized():
            self._rank = torch.distributed.get_rank()
        else:
            self._rank = int(os.environ.get("RANK", "0"))

        self._local_rank = self._rank % devices_per_node
        self._cur_step = int(os.environ["PROCESSED_ITERS"])

        # IPC Channels
        self._ipc_channels: Dict[IpcCommPurpose, IpcChannel] = {
            k: get_default_ipc_channel(purpose=k, local_rank=self._local_rank) for k in IpcCommPurpose
        }
        for ipc_channel in self._ipc_channels.values():
            ipc_channel.open()

        # Start a thread waiting for emergency save request.
        self._wait_emergency_save_thread = Thread(target=self._wait_for_emergency_save_request, daemon=True)
        self._wait_emergency_save_thread.start()

    @property
    def _is_step_started(self) -> bool:
        return self._step_start_time is not None

    @property
    def _has_locally_logged(self) -> bool:
        if self._log_path is None:
            return False

        return self._log_path.exists()

    def _teardown(self) -> None:
        """ Clean up resources.
        """
        for ipc_channel in self._ipc_channels.values():
            ipc_channel.close()
            ipc_channel.remove()

    def _wait_for_emergency_save_request(self) -> None:
        """ Wait for the emergency save request from the IPC channel.
        Do nothing for local mode.
        """
        msg = ipc_read(self._ipc_channels[IpcCommPurpose.EMERGENCY_SAVE])
        self._emergency_save_step = msg['emergency_save_step']

    def _local_log(self, msg):
        mode = "a" if self._has_locally_logged else "w"
        with self._log_path.open(mode=mode) as log_file:
            log_file.write(f"{json.dumps(msg)}\n")

    def init(self,
             total_train_steps: Optional[int] = None,
             local_log_name: Optional[str] = None) -> None:
        """Initialize the training manager.

        Args:
            total_train_steps: The number of total training steps
            local_log_name: log file name for local mode (only for local mode)

        Raises:
            PeriFlowError: when total_train_steps is not an integer
        """
        if self._is_local:
            if local_log_name is not None:
                self._log_path = Path(local_log_name)
            else:
                if torch.distributed.is_initialized():
                    # To prevent path overlap among processes, we add rank at the end of the log file name.
                    rank = torch.distributed.get_rank()
                    self._log_path = Path(f"./periflow_trainer_{int(time.time())}_{rank}.log")
                else:
                    self._log_path = Path(f"./periflow_trainer_{int(time.time())}.log")
        else:
            self._cloud_init()

            if total_train_steps is not None and total_train_steps <= self._cur_step:
                raise PeriFlowError(
                    'total_train_steps should be greater than the current step, '
                    f'current step = {self._cur_step}, total train step = {total_train_steps}')

            ipc_write(self._ipc_channels[IpcCommPurpose.LAST_STEP], {
                "step": total_train_steps
            })

        self.has_initialized = True

    @check_initialized
    def start_step(self) -> None:
        """Start a new training step.

        Raises:
            PeriFlowError: when the previous step is not ended
        """
        if self._is_step_started:
            raise PeriFlowError(
                'Step already started. Maybe `end_step` is not called for the previous step?)')

        self._cur_step += 1
        self._step_start_time = time.monotonic()
        self._checkpoint_saved_cur_step = False

    @check_initialized
    def end_step(self) -> None:
        """Finish the current training step.

        Raises:
            PeriFlowError: when try to call `end_step` before calling `start_step`
            PeriFlowInternalError: IPC failed
        """
        if not self._is_step_started:
            raise PeriFlowError(
                'Cannot call `end_step` before calling `start_step`')

        step_time = time.monotonic() - self._step_start_time
        if not self._is_local:
            msg = {
                "step": self._cur_step,
                "step_time": step_time,
            }
            ipc_write(self._ipc_channels[IpcCommPurpose.STEP_INFO], msg)
            ipc_status_read(self._ipc_channels[IpcCommPurpose.ACK])

        self._step_start_time = None

    @contextmanager
    def train_step(self):
        """Context manager wrapper for `start_step` and `end_step`."""
        self.start_step()
        try:
            yield
        finally:
            self.end_step()

    @check_initialized
    def is_emergency_save(self) -> bool:
        """Check whether emergency save should be handled or not.
        If emergency save step is -1, it means "save right now".

        Returns: True if emergency save is required for now.
        """
        return self._emergency_save_step in (self._cur_step, -1)

    @check_initialized
    def metric(self, msg: Dict) -> None:
        """Log a key-value metric dict and send it to Periflow. `step` info is added if not exists.
        Args:
            msg: A key-value dict containing user-defined metrics.

        Returns: None
        """
        if self._is_local:
            self._local_log(msg)
        else:
            new_msg = msg.copy()

            new_msg["step"] = self._cur_step
            new_msg["rank"] = self._rank
            new_msg["local_rank"] = self._local_rank

            ipc_write(self._ipc_channels[IpcCommPurpose.METRIC], new_msg)

    @check_initialized
    def upload_checkpoint(self):
        """Trigger uploading the checkpoint of the current step.

        For the last step, this function is a blocking call
        """
        if self._is_local:
            return

        if "CKPT_DIR" not in os.environ:
            periflow_logger.warning(
                "`upload_checkpoint` does nothing because `output_checkpoint_dir` is not configured when job launched.")
            return

        if self._checkpoint_saved_cur_step:
            raise PeriFlowError(
                f"Checkpoint is already uploaded for the step {self._cur_step}. Did you missed using pf.train_step?"
            )

        save_type = SaveType.EMERGENCY if self.is_emergency_save() else SaveType.NORMAL

        msg = {
            "step": self._cur_step,
            "save_type": save_type.value,
            "trigger_time": datetime.now().isoformat()
        }

        ipc_write(self._ipc_channels[IpcCommPurpose.CKPT], msg)
        self._checkpoint_saved_cur_step = True


periflow = TrainingManager()



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Exceptions that can be raised during IPC
"""

class IpcException(Exception):
    pass


class IpcConnectionError(IpcException):
    pass


class IpcChannelNotOpenedError(IpcException):
    pass


class IpcChannelIOError(IpcException):
    pass



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Helper functions
"""

import functools
from enum import Enum


class SaveType(str, Enum):
    NORMAL = "NORMAL"
    EMERGENCY = "EMERGENCY"


def check_initialized(func):
    @functools.wraps(func)
    def wrapper(training_manager, *args, **kwargs):
        assert training_manager.has_initialized, f'{func.__name__} must be called after pf.init()!'
        return func(training_manager, *args, **kwargs)
    return wrapper


def ensure_divisibility(numerator: int, denominator: int):
    assert numerator % denominator == 0, f"{numerator} is not divisible by {denominator}"



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""Wrapped errors for PeriFlow sdk
"""

class PeriFlowError(Exception):
    pass


class PeriFlowInternalError(PeriFlowError):
    def __init__(self, message=None):
        if message is None:
            message = "Please contact to FriendliAI"
        else:
            message = message.rstrip() + "\nPlease contant to FriendliAI"

        super().__init__(message)



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

""" IPC utils
"""

import asyncio
import errno
import json
import os
import struct
from enum import Enum
from typing import Optional

from periflow_sdk.comm.errors import (
    IpcChannelIOError,
    IpcChannelNotOpenedError,
    IpcConnectionError,
)
from periflow_sdk.errors import PeriFlowInternalError


class CommResultStatus(str, Enum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class IpcCommPurpose(str, Enum):
    STEP_INFO = "STEP_INFO"
    METRIC = "METRIC"
    ACK = "ACK"
    EMERGENCY_SAVE = "EMERGENCY_SAVE"
    LAST_STEP = "LAST_STEP"
    CKPT = "CKPT"


class FifoBase:
    """Abstraction for FIFO
    """
    def __init__(self, fifoname: str):
        self._fifo = None
        self._fifoname = fifoname
        _try_mkfifo(fifoname)

    @property
    def mode(self):
        raise NotImplementedError

    def open(self):
        if not self._fifo:
            self._fifo = os.open(self._fifoname, self.mode)

    def close(self):
        if self._fifo:
            try:
                os.close(self._fifo)
            except BrokenPipeError:
                pass
            self._fifo = None

    def remove(self):
        if self._fifo is not None:
            raise IpcChannelIOError("You should close the file first with FifoBase.close().")

        if os.path.exists(self._fifoname):
            os.remove(self._fifoname)


class FifoReader(FifoBase):
    """ Abstraction for FIFO reader side
    """
    @property
    def mode(self):
        # Read-only and non-blocking
        return os.O_RDONLY | os.O_NONBLOCK

    async def read(self) -> bytes:
        """Implementation of reading msg from the named pipe
        """
        # NOTE (TB): not compatible with Window OS
        # https://docs.python.org/3.8/library/asyncio-platforms.html#asyncio-platform-support
        assert self._fifo is not None
        loop = asyncio.get_running_loop()
        msg = bytearray()
        msg_size = None

        def _read(future: asyncio.Future):
            nonlocal msg_size
            try:
                if msg_size is None:
                    # read size first
                    msg_size_bytes = os.read(self._fifo, 4)
                    msg_size = _decode_msg_size(msg_size_bytes)

                msg_chunk = os.read(self._fifo, msg_size)
            except OSError as exc:
                if exc.errno == errno.EAGAIN:
                    future.set_result(b"")
                else:
                    future.set_exception(IpcConnectionError(str(exc)))
            except Exception as exc:  # pylint: disable=broad-except
                future.set_exception(IpcConnectionError(str(exc)))
            else:
                future.set_result(msg_chunk)
            finally:
                loop.remove_reader(self._fifo)

        while msg_size is None or msg_size > 0:
            future = loop.create_future()
            loop.add_reader(self._fifo, _read, future)
            msg_chunk = await future
            msg += msg_chunk
            msg_size -= len(msg_chunk)

        return bytes(msg)


class FifoWriter(FifoBase):
    """ Abstraction for FIFO writer side
    """
    @property
    def mode(self):
        return os.O_WRONLY | os.O_NONBLOCK

    async def write(self, content: bytes):
        """Implementation of writing msg into named pipe
        """
        assert self._fifo is not None
        loop = asyncio.get_running_loop()
        written_size = 0
        msg = _create_msg(content)
        msg_size = len(msg)

        def _write(future: asyncio.Future, m: bytes):
            try:
                wrote = os.write(self._fifo, m)
            except OSError as exc:
                if exc.errno == errno.EAGAIN:
                    future.set_result(0)
                else:
                    future.set_exception(IpcConnectionError(str(exc)))
            except Exception as exc:  # pylint: disable=broad-except
                future.set_exception(IpcConnectionError(str(exc)))
            else:
                future.set_result(wrote)
            finally:
                loop.remove_writer(self._fifo)

        while written_size < msg_size:
            future = loop.create_future()
            loop.add_writer(self._fifo, _write, future, msg)
            chunk_size = await future
            written_size += chunk_size
            msg = msg[chunk_size:]


class IpcChannel:
    """IPC communication channel
    """
    def __init__(self,
                 fifoname: str,
                 local_rank: Optional[int] = None):
        self._local_rank = local_rank
        self._fifoname = fifoname
        self._reader = FifoReader(fifoname)
        self._writer = FifoWriter(fifoname)
        self._opened = False

    async def read(self) -> dict:
        if not self._opened:
            error_msg = "IPC channel is not open. Call open() or __enter__ first."
            raise IpcChannelNotOpenedError(error_msg)

        msg = await self._reader.read()
        return json.loads(msg.decode())

    async def write(self, msg: dict):
        if not self._opened:
            error_msg = "IPC channel is not open. Call open() or __enter__ first."
            raise IpcChannelNotOpenedError(error_msg)

        bytes_msg = json.dumps(msg).encode()
        await self._writer.write(bytes_msg)

    def open(self):
        if self._opened:
            raise IpcChannelIOError("IPC channel is already open.")

        # NOTE: FIFO opening order
        # A process can open a FIFO in nonblocking mode. In this case, opening for read-only succeeds even if no one
        # has opened on the write side yet and opening for write-only fails with ENXIO (no such device or address)
        # unless the other end has already been opened.
        self._reader.open()
        self._writer.open()
        self._opened = True

    def close(self):
        if not self._opened:
            msg = "IPC channel is not open."
            raise IpcChannelNotOpenedError(msg)

        self._reader.close()
        self._writer.close()
        self._opened = False

    def remove(self):
        if self._opened:
            raise IpcChannelIOError(f"IPC channel is not closed yet: {self._fifoname}")
        self._reader.remove()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def fifoname(self) -> str:
        return self._fifoname

    @property
    def local_rank(self) -> Optional[int]:
        return self._local_rank

    @property
    def closed(self) -> bool:
        return not self._opened


def _encode_msg_size(size: int) -> bytes:
    """ Return a bytes object encoding the size of message.
    """
    return struct.pack("<I", size)


def _decode_msg_size(size_bytes: bytes) -> int:
    """ Return a message size in the integer format.
    """
    return struct.unpack("<I", size_bytes)[0]


def _create_msg(content: bytes) -> bytes:
    """ Create a message with the following format:

    ┌----------------┬--------------------------------┐
    | size (4 bytes) |        content (N bytes)       |
    └----------------┴--------------------------------┘
    """
    size = len(content)
    return _encode_msg_size(size) + content


def _try_mkfifo(fifoname: str):
    """ Create a FIFO (a named pipe) named path.
    """
    try:
        os.mkfifo(fifoname)
    except FileExistsError:
        pass


def get_default_ipc_channel(purpose: IpcCommPurpose, local_rank: int) -> IpcChannel:
    """ Create and return a IPC channel by the purpose.
    """
    if purpose == IpcCommPurpose.STEP_INFO:
        fifoname = f"/tmp/periflow_step_info_ipc_fifo_{local_rank}"
    elif purpose == IpcCommPurpose.METRIC:
        fifoname = f"/tmp/periflow_metric_ipc_fifo_{local_rank}"
    elif purpose == IpcCommPurpose.ACK:
        fifoname = f"/tmp/periflow_ack_ipc_fifo_{local_rank}"
    elif purpose == IpcCommPurpose.EMERGENCY_SAVE:
        fifoname = f"/tmp/periflow_emergency_save_ipc_fifo_{local_rank}"
    elif purpose == IpcCommPurpose.LAST_STEP:
        fifoname = f"/tmp/periflow_last_step_ipc_fifo_{local_rank}"
    elif purpose == IpcCommPurpose.CKPT:
        fifoname = f"/tmp/periflow_ckpt_ipc_fifo_{local_rank}"
    else:
        raise ValueError(f"Invalid purpose ({purpose}) is provided")
    return IpcChannel(fifoname, local_rank)


def ipc_read(channel: IpcChannel) -> dict:
    try:
        return asyncio.run(channel.read())
    except (IpcConnectionError, IpcChannelNotOpenedError) as exc:
        raise PeriFlowInternalError("IPC connection between PeriFlow and TrainingManager is Broken.") from exc


def ipc_status_read(channel: IpcChannel) -> dict:
    msg = ipc_read(channel)
    if msg["status"] != CommResultStatus.SUCCESS:
        raise PeriFlowInternalError(f"ERROR from PeriFlow: {msg}")
    return msg


def ipc_write(channel: IpcChannel, msg: dict) -> None:
    try:
        asyncio.run(channel.write(msg))
    except (IpcConnectionError, IpcChannelNotOpenedError) as exc:
        raise PeriFlowInternalError("IPC connection between PeriFlow and TrainingManager is Broken.") from exc



# Copyright (c) 2022-present, FriendliAI Inc. All rights reserved.

"""PeriFlow
"""

from periflow_sdk.manager import periflow

start_step = periflow.start_step
end_step = periflow.end_step
init = periflow.init
train_step = periflow.train_step
is_emergency_save = periflow.is_emergency_save
metric = periflow.metric
upload_checkpoint = periflow.upload_checkpoint

__all__ = [
    'start_step',
    'end_step',
    'init',
    'train_step',
    'is_emergency_save',
    'metric',
    'upload_checkpoint',
]



# Copyright (C) 2021 friendli.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
from setuptools import setup, find_packages


def read(fname):
    """
    Args:
        fname:
    """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def read_version():
    return read("VERSION").strip()


COMMON_DEPS = [
]

TEST_DEPS = [
    "coverage==5.5",
    "pytest==6.2.4",
    "pytest-asyncio==0.15.1",
    "pytest-cov==2.11.1",
    "pytest-benchmark==3.4.1",
    "pytest-lazy-fixture==0.6.3",
]

setup(
    name='periflow_sdk',
    version=read_version(),
    author = 'FriendliAI',
    license="Apache License 2.0",
    url = "https://github.com/friendliai/periflow-python-sdk",
    description = "PeriFlow SDK",
    packages=find_packages(include=['periflow_sdk', 'periflow_sdk.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
    ],
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=COMMON_DEPS,
    extras_require={
        "test": TEST_DEPS,
    }
)
